import "dart:math" as m;
import "./agent.dart";
import "./utils.dart" as utils;
import "./exception.dart" as exceptions;
import 'property.dart';
import 'propertyWithDeviation.dart';

/// An Agent encapsulates a list of properties
class Agent {
  /// Constructor creates an Agent from a List of Properties.
  Agent(List<Property> this._vector);

  /// Create Agent with all properties having numerically the same center and deviation  nur mit Strategie
  factory Agent.createUniformWithDeviation(
      double center, double deviation, int size) {
    // Der Suchraum sollte zu Iteration bei der Generierung dienen.
    List<Property> vector = new List<PropertyWithDeviation>();
    // for size of search space
    for (var i = 0; i < size; i++) {
      var p = new PropertyWithDeviation(center, deviation);
      vector.add(p);
      print("create ${p}");
    }
    return new Agent(vector);
  }

  /// Create an Agen from a List of simple values and deviations (strategies)

  factory Agent.createFromVectorWithDeviation(
      List<double> values, List<double> deviations) {
    int fl = values.length;
    int sl = deviations.length;

    List<Property> vector = new List<PropertyWithDeviation>();

    if (fl == sl) {
      for (var i = 0; i < sl; i++) {
        var p = new PropertyWithDeviation(0, 1);
        vector.add(p);
      }
    } else {
      throw new exceptions.RuntimeException(
          "Agent.createFromVector - Lengths differ: center ${fl} vs. deviation ${sl}");
    }
    return new Agent(vector);
  }

  ///Create a normal (0,1)-Agent from length
  factory Agent.createNormalWithDeviation(int size) {
    List<Property> vector = new List<PropertyWithDeviation>();
    for (var i = 0; i < size; i++) {
      var p = new PropertyWithDeviation(0.0, 1.0);
      vector.add(p);
    }
    return new Agent(vector);
  }

  /// Create a normal (0,1)-Agent from  a crossover
  /// values and deviations a derived from parents
  factory Agent.createFromCrossover(Agent first, Agent second) {
    int fl = first.length;
    int sl = second.length;

    List<PropertyWithDeviation> cv = new List<PropertyWithDeviation>();
    if (fl == sl) {
      List<PropertyWithDeviation> fv = first.vector;
      List<PropertyWithDeviation> sv = second.vector;
      List<PropertyWithDeviation> cv;
      for (int i = 0; i < fl; i++) {
        PropertyWithDeviation fg = fv[i];
        PropertyWithDeviation sg = sv[i];
        cv.add(new PropertyWithDeviation(
            (fg.value + sg.value) / 2.0, (fg.deviation + sg.deviation) / 2.0));
      }
    } else {
      throw new exceptions.RuntimeException(
          "Crossover - Lengths differ: ${fl} vs ${sl}");
    }
    return new Agent(cv);
  }
  // functional-style Mutation of Agent
  /* 
  factory Agent.createFromMutation(Agent parent) {
    List<PropertyWithDeviation> pv = parent.vector;
    List<PropertyWithDeviation> cv = new  List<PropertyWithDeviation>();
    for (int i = 0; i < pv.length; i++) {
      PropertyWithDeviation pg = pv[i];
      cv.add(new PropertyWithDeviation(pg.value + utils.rdouble() * pg.deviation,
          pg.deviation + utils.rdouble()));

    }
    return new Agent(cv);
  }
  */

  // egal ob mit Strategie
  factory Agent.createFromMutation(Agent parent) {
    var c = new Agent.createAsClone(parent);
    c.mutate();
    return c;
  }

  /// Create an Agent by cloning, copy
  factory Agent.createAsClone(Agent parent) {
    List<PropertyWithDeviation> pv = parent.vector;
    List<PropertyWithDeviation> cv = new List<PropertyWithDeviation>();
    for (int i = 0; i < pv.length; i++) {
      PropertyWithDeviation pg = pv[i];
      cv.add(new PropertyWithDeviation(pg.value, pg.deviation));
    }
    return new Agent(cv);
  }
  Agent clone() => Agent.createAsClone(this);

  bool equals(Agent other) {
    List<PropertyWithDeviation> ov = other.vector;
    for (int i = 0; i < ov.length; i++) {
      if (this.vector[i] != ov[i]) return false;
    }
    return true;
  }

  /// Vector of properties.
  List<Property> _vector;

  List<Property> get vector => this._vector;
  set vector(List<Property> v) => this._vector = v;

  /// Number of properties
  int get length => this._vector.length;

  double objective_function(List<PropertyWithDeviation> vector) {
    return vector.fold(0.0, (sum, x) => sum + (m.pow(x.value, 2.0)));
  }

  /// Fitnes as a Number
  double fitness = 0;

  /// Evaluate properties into fitness
  void evaluate() {
    this.fitness = this.objective_function(this.vector);
  }

  /// Mutation of values
  void mutate() {
    vector.forEach((Property x) => x.mutate());
  }

  @override
  String toString() {
    return ("Agent " +
        "Fitness: " +
        this.fitness.toString() +
        " Vector: " +
        this.vector.toString());
  }
}
