import 'dart:math' as m;
import 'property.dart';

/**
 * Property encapsulates a Value and a deviation
 * 
 */
class PropertyWithDeviation extends Property {
  ///
  PropertyWithDeviation(double this._value, double this._deviation)
      : super(_value);

  factory PropertyWithDeviation.simpleCreate() {
    return new PropertyWithDeviation(0, 1);
  }

  /// Create a property by cross over of two properties.
  factory PropertyWithDeviation.createFromCrossover(
      PropertyWithDeviation first, PropertyWithDeviation second) {
    var value = (first.value + second.value) / 2.0;
    var deviation = (first.deviation + second.deviation) / 2.0;
    return new PropertyWithDeviation(value, deviation);
  }

  /// The encapsulated value

  double _value;

  @override
  m.Random r = new m.Random();

  @override
  double get value => this._value;

  @override
  set value(double v) => this._value = v;

  // The encapsulated deviation
  double _deviation;
  double get deviation => this._deviation;
  set deviation(double d) => this._deviation = d;

  @override
  String toString() {
    return """Value  ${this.value} Deviation ${this.deviation} """;
  }

  /// Mutation of deviation
  double mutateDeviation(double dev) {
    double res = dev * m.exp(r.nextDouble() + r.nextDouble());
    return res;
  }

  /// Mutation of value using deviation
  double mutateProblemWithDeviation(double vector, double dev) {
    m.Random r = new m.Random();
    double res = vector + dev * ((r.nextDouble() - 0.5) * 2.0);
    return res;
  }

  /// Mutation of value using strategy and then mutation of deviation
  @override
  void mutate() {
    var r = mutateProblemWithDeviation(this.value, this.deviation);
    this.value = r;
    var d = mutateDeviation(this.deviation);
    this.deviation = d;
  }
}
